/*
 * Copyright (C) 2009  David I. Bell
 *
 * Calc is open software; you can redistribute it and/or modify it under
 * the terms of the version 2.1 of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Calc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General
 * Public License for more details.
 *
 * A copy of version 2.1 of the GNU Lesser General Public License is
 * distributed with calc under the filename COPYING-LGPL.  You should have
 * received a copy with calc; if not, write to Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Node calculation module.
 *
 * This executes those nodes which are used for expression evaluation.
 * The arguments for the nodes are either contained as special data
 * defined for the node (such as literal constants), or else which are
 * generated from the kid nodes of a node.  An evaluation stack is used
 * to provide the arguments from the kid nodes, and a node's result is
 * pushed back onto the evaluation stack.
 *
 * The evaluation of a node is performed by recursively walking down
 * the node tree starting with the indicated node, and evaluating the
 * values of each of the kid nodes, and then combining those results.
 * A slight exception to this is performed by the alternate, logical
 * OR, and logical AND nodes which selectively evaluate their kid nodes.
 *
 * User defined function calls are executed by pushing the arguments onto
 * the evaluation stack, and then calling a calculate routine to execute
 * the function.  The evaluation of the function is done using its own
 * evaluation and execution stacks and so doesn't otherwise affect us.
 * The function value is returned by the call.
 *
 * This module contains no program flow nodes.  None of the nodes here
 * affect the execution of statements for a function unless a math error
 * is generated.  So these methods make no use of the execution stack
 * (except indirectly from a user-defined function call).
 */

#include "calc.h"
#include "node.h"
#include "nodeinfo.h"
#include "userfunc.h"
#include "builtinfunc.h"
#include "symbol.h"
#include "hist.h"
#include "args.h"
#include "file.h"
#include "zrand.h"
#include "have_fpos.h"
#include "string.h"
#include "method.h"


/*
 * A null value.
 */
static	VALUE	nullValue = {V_NULL, V_NOSUBTYPE};


/*
 * Save the value of our kid as a value history and old value
 * (if it is not null) and also push it onto the evaluation stack.
 */
void
ocSaveHistory(const Node * node)
{
	VALUE *	v1;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);

	if (v1->v_type != V_NULL)
	{
		copyvalue(v1, &tempValue);

		SaveOldValue(&tempValue);
	}
}


/*
 * Save the value of our kid as the default return value and also
 * push it onto the evaluation stack.
 */
void
ocSaveReturn(const Node * node)
{
	VALUE *	v1;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);

	copyvalue(v1, &tempValue);

	SaveReturnValue(&tempValue);
}


/*
 * Push the address of the object value for a method onto the
 * evaluation stack.
 */
void
ocThisAddr(const Node * node)
{
	VALUE *	thisValue;
	VALUE *	result;

	thisValue = GetThisValue();

	if (thisValue == 0)
		math_error("Null \"this\" value");

	result = NewValue();
	result->v_addr = thisValue;
	result->v_type = V_ADDR;
}


/*
 * Push the address of the object's SUPER value for a method onto the
 * evaluation stack.  This is currently the same as the 'this' value.
 */
void
ocSuperAddr(const Node * node)
{
	VALUE *	superValue;
	VALUE *	result;

	superValue = GetThisValue();

	if (superValue == 0)
		math_error("Null \"super\" value");

	result = NewValue();
	result->v_addr = superValue;
	result->v_type = V_ADDR;
}


/*
 * Push the address of a local variable onto the evaluation stack.
 */
void
ocLocalAddr(const Node * node)
{
	VALUE *	localValue;
	VALUE *	result;

	/*
	 * Get the address of the local variable.
	 */
	localValue = GetLocalVariable(node->id);

	if (localValue == 0)
	{
		math_error("Bad local variable index");
		/*NOTREACHED*/
	}

	/*
	 * Push a value which points to the appropriate local variable.
	 */
	result = NewValue();
	result->v_addr = localValue;
	result->v_type = V_ADDR;
}


/*
 * Push the address of a global variable onto the evaluation stack.
 */
void
ocGlobalAddr(const Node * node)
{
	VALUE *	result;
	GLOBAL *sp;

	sp = (GLOBAL *) node->data;

	if (sp == 0)
	{
		math_error("Global variable not initialized");
		/*NOTREACHED*/
	}

	result = NewValue();
	result->v_addr = &sp->g_value;
	result->v_type = V_ADDR;
}


/*
 * Push the address of a function parameter variable onto the
 * evaluation stack.
 */
void
ocParamAddr(const Node * node)
{
	VALUE *	paramValue;
	VALUE *	result;

	/*
	 * Get the argument's address.
	 */
	paramValue = GetArgumentVariable(node->id);

	if (paramValue == 0)
	{
		math_error("Bad parameter index");
		/*NOTREACHED*/
	}

	/*
	 * Push the argument address onto the evaluation stack.
	 */
	result = NewValue();

	if (paramValue->v_type == V_ADDR)
		result->v_addr = paramValue->v_addr;
	else
		result->v_addr = paramValue;

	result->v_type = V_ADDR;
}


/*
 * Push the value of a local variable onto the evaluation stack.
 */
void
ocLocalValue(const Node * node)
{
	VALUE *	localValue;
	VALUE	result;

	/*
	 * Get the address of the local variable.
	 */
	localValue = GetLocalVariable(node->id);

	if (localValue == 0)
	{
		math_error("Bad local variable index");
		/*NOTREACHED*/
	}

	/*
	 * Push the value of the local variable.
	 */
	copyvalue(localValue, &result);

	PushValue(&result);
}


/*
 * Push the value of a global variable onto the evaluation stack.
 */
void
ocGlobalValue(const Node * node)
{
	GLOBAL *sp;
	VALUE	result;

	sp = (GLOBAL *) node->data;

	if (sp == 0)
	{
		math_error("Global variable not initialized");
		/*NOTREACHED*/
	}

	copyvalue(&sp->g_value, &result);

	PushValue(&result);
}


/*
 * Push the value of a function parameter variable onto the
 * evaluation stack.
 */
void
ocParamValue(const Node * node)
{
	VALUE *	paramValue;
	VALUE	result;

	/*
	 * Get the argument's address.
	 */
	paramValue = GetArgumentVariable(node->id);

	if (paramValue == 0)
	{
		math_error("Bad parameter index");
		/*NOTREACHED*/
	}

	/*
	 * Push the argument address onto the evaluation stack.
	 */
	if (paramValue->v_type == V_ADDR)
		paramValue = paramValue->v_addr;

	copyvalue(paramValue, &result);

	PushValue(&result);
}


/*
 * Push the linked value of a local variable onto the evaluation stack.
 */
void
ocLocalLink(const Node * node)
{
	VALUE *	localValue;
	VALUE	result;

	/*
	 * Get the address of the local variable.
	 */
	localValue = GetLocalVariable(node->id);

	if (localValue == 0)
	{
		math_error("Bad local variable index");
		/*NOTREACHED*/
	}

	/*
	 * Push the linked value of the local variable.
	 */
	LinkValue(localValue, &result);

	PushValue(&result);
}


/*
 * Push the linked value of a global variable onto the evaluation stack.
 */
void
ocGlobalLink(const Node * node)
{
	GLOBAL *sp;
	VALUE	result;

	sp = (GLOBAL *) node->data;

	if (sp == 0)
	{
		math_error("Global variable not initialized");
		/*NOTREACHED*/
	}

	/*
	 * Push the linked value of the global variable.
	 */
	LinkValue(&sp->g_value, &result);

	PushValue(&result);
}


/*
 * Push the linked value of a function parameter variable onto the
 * evaluation stack.
 */
void
ocParamLink(const Node * node)
{
	VALUE *	paramValue;
	VALUE	result;

	/*
	 * Get the argument's address.
	 */
	paramValue = GetArgumentVariable(node->id);

	if (paramValue == 0)
	{
		math_error("Bad parameter index");
		/*NOTREACHED*/
	}

	if (paramValue->v_type == V_ADDR)
		paramValue = paramValue->v_addr;

	/*
	 * Push the linked argument value onto the evaluation stack.
	 */
	LinkValue(paramValue, &result);

	PushValue(&result);
}


/*
 * Push the value of a function parameter onto the evaluation stack.
 * Our kid's value is the index of the argument number (starting at 1).
 * A value of 0 returns the number of arguments.
 */
void
ocArgValue(const Node * node)
{
	VALUE *	v1;
	VALUE *	argValue;
	long index;

	v1 = CalcValue(node->kids[0]);

	if ((v1->v_type != V_NUM) || qisneg(v1->v_num) ||
		qisfrac(v1->v_num))
	{
		math_error("Illegal argument for arg function");
		/*NOTREACHED*/
	}

	/*
	 * If the argument is zero then return the number of arguments.
	 */	 
	if (qiszero(v1->v_num))
	{
		PopValue();
		PushNumber(itoq((long) GetArgumentCount()));

		return;
	}

	/*
	 * The argument is positive.
	 * Extract its value for use in indexing.
	 */
	index = qtoi(v1->v_num) - 1;

	PopValue();

	/*
	 * Get the argument's address.
	 * If the argument is unknown then push the undefined value.
	 */
	argValue = GetArgumentVariable(index);

	if (argValue == 0)
	{
		math_error("Bad parameter index");
		/*NOTREACHED*/
	}

	if (argValue->v_type == V_ADDR)
		argValue = argValue->v_addr;

	PushValue(argValue);
}


/*
 * Push an imaginary constant onto the evaluation stack.
 */
void
ocImaginary(const Node * node)
{
	VALUE *		result;
	NUMBER *	q;
	COMPLEX *	com;

	q = constvalue(node->id);

	if (q == NULL)
	{
		math_error("Numeric constant value not found");
		/*NOTREACHED*/
	}

	if (qiszero(q))
	{
		PushNumber(qlink(q));

		return;
	}

	com = comalloc();
	com->real = qlink(&_qzero_);
	com->imag = qlink(q);

	result = NewValue();
	result->v_com = com;
	result->v_type = V_COM;
}


/*
 * Create a matrix whose sizes are given by our kids and push it
 * onto the evaluation stack.  Each pair of kids specifies the minimum
 * and maximum index values for a dimension.
 */
void
ocMatrixCreate(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE *	vp;
	VALUE *	result;
	MATRIX *mat;		/* matrix being defined */
	NUMBER *num1;		/* first number from stack */
	NUMBER *num2;		/* second number from stack */
	long	min[MAXDIM];	/* minimum range */
	long	max[MAXDIM];	/* maximum range */
	long	index;		/* index */
	long	tmp;		/* temporary */
	long	size;		/* size of matrix */
	int	dim;		/* dimension */

	/*
	 * The dimension of the matrix is half of the number of kids.
	 */
	if ((node->count & 0x01) != 0)
	{
		math_error("Odd kid count for matrix");
		/*NOTREACHED*/
	}

	dim = node->count / 2;

	if ((dim <= 0) || (dim > MAXDIM))
	{
		math_error("Bad dimension %ld for matrix", dim);
		/*NOTREACHED*/
	}

	/*
	 * Evaluate all of the dimensions and calculate the final size.
	 */
	size = 1;

	for (index = dim - 1; index >= 0; index--)
	{
		v1 = CalcValue(node->kids[index * 2]);
		v2 = CalcValue(node->kids[index * 2 + 1]);

		if ((v1->v_type != V_NUM) || (v2->v_type != V_NUM))
		{
			math_error("Non-numeric bounds for matrix");
			/*NOTREACHED*/
		}

		num1 = v1->v_num;
		num2 = v2->v_num;

		if (qisfrac(num1) || qisfrac(num2))
		{
			math_error("Non-integral bounds for matrix");
			/*NOTREACHED*/
		}

		if (zge31b(num1->num) || zge31b(num2->num))
		{
			math_error("Very large bounds for matrix");
			/*NOTREACHED*/
		}

		min[index] = qtoi(num1);
		max[index] = qtoi(num2);

		if (min[index] > max[index])
		{
			tmp = min[index];
			min[index] = max[index];
			max[index] = tmp;
		}

		size *= (max[index] - min[index] + 1);

		if (size > 10000000)
		{
			math_error("Very large size for matrix");
			/*NOTREACHED*/
		}

		PopValue();
		PopValue();
	}

	/*
	 * Now allocate the matrix and fill in its details.
	 */
	mat = matalloc(size);
	mat->m_dim = dim;

	for (index = 0; index < dim; index++)
	{
		mat->m_min[index] = min[index];
		mat->m_max[index] = max[index];
	}

	/*
	 * Fill the matrix will zero values.
	 * Note: This knows that a new matrix has NULL element values
	 * and so no freeing of old values is required.
	 */
	vp = mat->m_table;

	for (index = 0; index < mat->m_size; index++, vp++)
	{
		vp->v_num = qlink(&_qzero_);
		vp->v_type = V_NUM;
		vp->v_subtype = V_NOSUBTYPE;
	}

	/*
	 * Push the matrix onto the stack.
	 */
	result = NewValue();

	result->v_mat = mat;
	result->v_type = V_MAT;
}


/*
 * Perform an initialization for some elements of a matrix or object.
 * Our kids are a pointer to a value to be initialized and one or more
 * values to initialize it with.  By default, the indexes of the elements
 * to be initialized starts with zero and increments by one for each
 * additional element.  However, any of the kids can be an integer literal
 * value which will change the index for the next element to the specified
 * value.  The original object is left on the evaluation stack.
 */
void
ocInitializeElements(const Node * node)
{
	Node *		kidNode;
	VALUE *		v1;
	VALUE *		v2;
	VALUE *		elementTable;
	VALUE *		elementValue;
	MATRIX *	mat;
	OBJECT *	obj;
	long		tableSize;
	long		elementIndex;
	int		kidIndex;

	elementTable = 0;
	tableSize = 0;

	/*
	 * Get the object to be initialized.
	 * This object will be left on the evaluation stack as the result.
	 */
	v1 = CalcValue(node->kids[0]);

	/*
	 * Get the base element table address and table size for the
	 * value to be initialized.
	 */
	switch (v1->v_type)
	{
		case V_MAT:
			mat = v1->v_mat;

			elementTable = mat->m_table;
			tableSize = mat->m_size;

			break;

		case V_OBJ:
			obj = v1->v_obj;

			elementTable = obj->o_table;
			tableSize = obj->o_actions->oa_count;

			break;

		default:
			math_error("Attempt to initialize non matrix or object");
			/*NOTREACHED*/
	}

	/*
	 * First range check the element indexes to be assigned into.
	 * We have to scan through the kids looking for any literal
	 * integer nodes which might reset the element index.
	 */
	elementIndex = 0;

	for (kidIndex = 1; kidIndex < node->count; kidIndex++)
	{
		kidNode = node->kids[kidIndex];

		/*
		 * If this is an integer literal node then set the
		 * element index to its id value, otherwise increment
		 * the element index if it is in range.
		 */
		if (kidNode->info->opcode == OP_INT_LITERAL)
			elementIndex = kidNode->id;
		else if ((elementIndex >= 0) && (elementIndex < tableSize))
			elementIndex++;
		else
		{
			math_error("Attempt to initialize element %ld",
				elementIndex);

			/*NOTREACHED*/
		}
	}

	/*
	 * The indexes are all valid.  Now walk through the kids again
	 * and actually perform the initialization of the elements.
	 */
	elementIndex = 0;

	for (kidIndex = 1; kidIndex < node->count; kidIndex++)
	{
		kidNode = node->kids[kidIndex];

		/*
		 * If this is an integer literal node then set the
		 * element index to its value and continue with the
		 * next kid.
		 */
		if (kidNode->info->opcode == OP_INT_LITERAL)
		{
			elementIndex = kidNode->id;

			continue;
		}

		/*
		 * This is a value to initialize with, so calculate it.
		 */
		v2 = CalcValue(node->kids[kidIndex]);

		/*
		 * Index into the table by the element index.
		 */
		elementValue = elementTable + elementIndex;

		/*
		 * Replace the old element value with the new one.
		 */
		freevalue(elementValue);
		copyvalue(v2, elementValue);

		/*
		 * Pop off the initialization value.
		 */
		PopValue();

		/*
		 * Advance to the next element index.
		 */
		elementIndex++;
	}
}


/*
 * Calculate the address of an index for a matrix, object, string or block.
 * The address can be be used for reading or writing according to whether
 * our id value is nonzero.  The kids are the index values.  The resulting
 * address is pushed onto the evaluation stack, and can either be a V_ADDR
 * value referencing a complete VALUE, or else a V_SUBADDR value to
 * reference a subvalue of a VALUE (such as an individual byte).
 */
void
ocIndexAddr(const Node * node)
{
	int	dim;
	long	index;
	BOOL	writeFlag;
	VALUE *	v1;
	VALUE *	vp;
	VALUE *	result;
	VALUE *	indices[MAXDIM];
	VALUE	resultValue;

	/*
	 * Get the write flag from the id value.
	 */
	writeFlag = (node->id != 0);

	/*
	 * Get the dimension count which is one less than the number
	 * of kids.
	 */
	dim = node->count - 1;

	if ((dim <= 0) || (dim > MAXDIM))
	{
		math_error("Bad number of dimensions for indexing");
		return;
	}

	/*
	 * Get the object which is to be indexed.
	 * Make sure that a full address was passed and then get the
	 * actual value to be indexed.
	 */
	v1 = CalcFullAddress(node->kids[0]);

	v1 = v1->v_addr;

	/*
	 * Build a table of all of the index values.
	 */
	for (index = 0; index < dim; index++)
		indices[index] = CalcValue(node->kids[index + 1]);

	/*
	 * Switch on the type of data to be indexed.
	 */
	switch (v1->v_type)
	{
		case V_MAT:
			vp = matindex(v1->v_mat, writeFlag, dim, indices);
			break;

		case V_ASSOC:
			vp = associndex(v1->v_assoc, writeFlag, dim, indices);

			if (vp == 0)
				vp = &nullValue;

			break;

		case V_STR:
			resultValue = StringGetIndexValue(v1, writeFlag,
				dim, indices);

			vp = &resultValue;

			break;

		case V_BLOCK:
		case V_NBLOCK:
			resultValue = BlockGetIndexValue(v1, writeFlag,
				dim, indices);

			vp = &resultValue;

			break;

		default:
			math_error("Illegal value type \"%s\" for indexing",
				valuetype(v1));

			/*NOTREACHED*/
			return;
	}

	/*
	 * Remove all of the kid's values.
	 */
	while (dim-- > 0)
		PopValue();

	/*
	 * Remove the original object too.
	 */
	PopValue();

	/*
	 * If an result value was returned then free the original
	 * value and push the new value.
	 */
	if (vp == &resultValue)
	{
		PushValue(vp);

		return;
	}

	/*
	 * A true index was produced.
	 * Push the address as the result.
	 */
	result = NewValue();

	result->v_addr = vp;
	result->v_type = V_ADDR;
}


/*
 * Push the address of an element of a matrix or object onto the
 * evaluation stack.  The index of the element is a constant encoded
 * into the node.
 */
void
ocElementAddr(const Node * node)
{
	VALUE *		v1;
	VALUE *		result;
	MATRIX *	mat;
	OBJECT *	obj;
	void *		address;
	int		index;
	int		offset;

	index = (int) node->id;

	/*
	 * Get the kid's value which must be a full address.
	 */
	v1 = CalcFullAddress(node->kids[0]);

	v1 = v1->v_addr;

	/*
	 * Switch on the type of object to index it.
	 */
	switch (v1->v_type)
	{
		case V_MAT:
			mat = v1->v_mat;

			if ((index < 0) || (index >= mat->m_size))
			{
				math_error("Non-existent element for matrix");
				return;
			}

			address = &mat->m_table[index];

			break;

		case V_OBJ:
			obj = v1->v_obj;
			offset = objoffset(obj, index);

			if (offset < 0)
			{
				math_error("Non-existent element for object");
				return;
			}

			address = &obj->o_table[offset];

			break;

		default:
			math_error("Cannot reference element for value type \"%s\"",
				valuetype(v1));

			return;
	}

	/*
	 * Remove the kid's value and replace it with the address of
	 * the element.
	 */
	PopValue();

	result = NewValue();
	result->v_addr = address;
	result->v_type = V_ADDR;
}


/*
 * Push the value of an element of a matrix or object onto the
 * evaluation stack.  The index of the element is a constant encoded
 * into the node.
 *
void
ocElementValue(const Node * node)
{
	VALUE *	v1;

	ocElemAddr(node);

	v1 = TopValue();

	v1 = v1->v_addr;

	PopValue();
	PushValue(v1->v_addr);
}


/*
 * Allocate an instance of an object and push it onto the evaluation stack.
 * The type of object to create is a constant encoded into the node.
 */
void
ocObjectCreate(const Node * node)
{
	VALUE *	result;

	result = NewValue();

	result->v_obj = objalloc(node->id);
	result->v_type = V_OBJ;
}


/*
 * Initialize the value of an object and push the object onto the
 * evaluation stack.  (Actually, the object is already on the stack
 * so we just leave it there.)
 */
void
ocObjectInitialize(const Node * node)
{
	VALUE *	v1;

	v1 = CalcValue(node->kids[0]);

	(void) ObjectCall(OBJ_INIT, v1, NULL_VALUE, NULL_VALUE);
}


/*
 * Assign the result of an expression into a static variable once and
 * push the address of the variable onto the evaluation stack.
 * After the first assignment has been performed, this node will not
 * reevaluate or assign the expression.
 */
void
ocInitStatic(const Node * node)
{
	Node *		varNode;
	GLOBAL *	sp;
	VALUE *		v1;
	VALUE *		v2;
	VALUE *		var;
	VALUE		dataValue;

	/*
	 * Get the static variable value to be initialized.
	 * Our first kid node must be a GLOBAL_ADDR.
	 */
	varNode = node->kids[0];

	if (varNode->info->opcode != OP_GLOBAL_ADDR)
	{
		math_error("Invalid InitStatic kid");
		/*NOTREACHED*/
	}

	/*
	 * Get the global variable.
	 */
	sp = (GLOBAL *) varNode->data;

	if (sp == 0)
	{
		math_error("Global variable not initialized");
		/*NOTREACHED*/
	}

	/*
	 * Push the address of the variable's value onto the evaluation
	 * stack since we need to return it in any case.
	 */
	v1 = NewValue();
	v1->v_addr = &sp->g_value;
	v1->v_type = V_ADDR;

	/*
	 * If the variable is already initialized then do nothing else.
	 */
	if (sp->g_staticInited)
		return;

	/*
	 * This is the first initialization of the variable.
	 * So proceed like a normal assignment.
	 */
	var = v1->v_addr;
	v2 = CalcRawValue(node->kids[1]);

	/*
	 * Switch on the value type we are initializing with.
	 */
	switch (v2->v_type)
	{
		case V_ADDR:
			/*
			 * The second argument is the address of a
			 * variable so we need to copy it's value.
			 */
			v2 = v2->v_addr;

			PopValue();

			/*
			 * If we are assigning to ourself then do nothing.
			 */
			if (v2 == var)
			{
				sp->g_staticInited = TRUE;

				return;
			}

			/*
			 * Copy the value of the variable into our temporary.
			 */
			copyvalue(v2, &dataValue);

			break;

		case V_SUBADDR:
			/*
			 * The second argument is a sub-value so we need
			 * to extract the value.
			 */
			dataValue = LoadValueFromAddress(v2);

			break;

		default:
			/*
			 * The second value is a real stack value.
			 * Extract it from the stack into our temporary.
			 */
			ExtractValue(&dataValue);
	}

	/*
	 * Free the variable's value and copy in the new one.
	 */
	freevalue(var);
	*var = dataValue;

	/*
	 * Set the inited flag so we won't do this again.
	 */
	sp->g_staticInited = TRUE;
}


/*
 * Assign the result of an expression into an address and push the
 * address back onto the evaluation stack.  The address can either
 * be a V_ADDR referencing a complete VALUE to be replaced or else a
 * V_SUBADDR referencing just a component of a VALUE.
 */
void
ocAssign(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	dataValue;

	/*
	 * Calculate the address to assign into and the raw value
	 * to be assigned,
	 */
	v1 = CalcAddress(node->kids[0]);
	v2 = CalcRawValue(node->kids[1]);

	/*
	 * Switch on the raw second value to collect its value.
	 * We must do this manually here in case we are assigning a
	 * variable into itself, otherwise a CalcValue call could have
	 * been used above instead.
	 */
	switch (v2->v_type)
	{
		case V_ADDR:
			/*
			 * The value is contained in an address.
			 * We need to copy the value from the address.
			 */
			v2 = v2->v_addr;

			PopValue();

			/*
			 * If we are assigning a value into itself then
			 * do nothing.
			 */
			if ((v1->v_type == V_ADDR) && (v1->v_addr == v2))
				return;

			/*
			 * Link the value from the address into our temporary.
			 */
			LinkValue(v2, &dataValue);

			break;

		case V_SUBADDR:
			/*
			 * The value is in a subaddress.
			 * Load the subvalue from the value.
			 */
			dataValue = LoadValueFromAddress(v2);

			PopValue();

			break;

		default:
			/*
			 * The second value is a value on the stack.
			 * Extract it from the stack into our temporary.
			 */
			ExtractValue(&dataValue);

			break;
	}

	/*
	 * If we are assigning into an object value then call a user
	 * defined function to do that.
	 */
	if ((v1->v_type == V_ADDR) && (v1->v_addr->v_type == V_OBJ))
	{
		(void) ObjectCall(OBJ_ASSIGN, v1->v_addr,
			&dataValue, NULL_VALUE);

		return;
	}

	/*
	 * Store the value into the address.
	 */
	StoreValueIntoAddress(v1, &dataValue);
}


/*
 * Assign a zero value to a variable and push the variable onto the
 * evaluation stack.
 */
void
ocAssignZero(const Node * node)
{
	VALUE *	v1;
	VALUE	dataValue;

	/*
	 * Calculate the address to be assigned.
	 */
	v1 = CalcAddress(node->kids[0]);

	/*
	 * Get a zero value.
	 */
	dataValue.v_num = qlink(&_qzero_);
	dataValue.v_type = V_NUM;
	dataValue.v_subtype = V_NOSUBTYPE;

	/*
	 * If we are assigning into an object value then call a user
	 * defined function to do that.
	 */
	if ((v1->v_type == V_ADDR) && (v1->v_addr->v_type == V_OBJ))
	{
		(void) ObjectCall(OBJ_ASSIGN, v1->v_addr,
			&dataValue, NULL_VALUE);

		qfree(dataValue.v_num);

		return;
	}

	/*
	 * Store the value into the address.
	 */
	StoreValueIntoAddress(v1, &dataValue);
}


/*
 * Assign a one value to a variable and push the variable onto the
 * evaluation stack.
 */
void
ocAssignOne(const Node * node)
{
	VALUE *	v1;
	VALUE	dataValue;

	/*
	 * Calculate the address to be assigned.
	 */
	v1 = CalcAddress(node->kids[0]);

	/*
	 * Get a one value.
	 */
	dataValue.v_num = qlink(&_qone_);
	dataValue.v_type = V_NUM;
	dataValue.v_subtype = V_NOSUBTYPE;

	/*
	 * If we are assigning into an object value then call a user
	 * defined function to do that.
	 */
	if ((v1->v_type == V_ADDR) && (v1->v_addr->v_type == V_OBJ))
	{
		(void) ObjectCall(OBJ_ASSIGN, v1->v_addr,
			&dataValue, NULL_VALUE);

		qfree(dataValue.v_num);

		return;
	}

	/*
	 * Store the value into the address.
	 */
	StoreValueIntoAddress(v1, &dataValue);
}


/*
 * Assign a minus one value to a variable and push the variable onto the
 * evaluation stack.
 */
void
ocAssignMinusOne(const Node * node)
{
	VALUE *	v1;
	VALUE	dataValue;

	/*
	 * Calculate the address to be assigned.
	 */
	v1 = CalcAddress(node->kids[0]);

	/*
	 * Get a minus one value.
	 */
	dataValue.v_num = qlink(&_qnegone_);
	dataValue.v_type = V_NUM;
	dataValue.v_subtype = V_NOSUBTYPE;

	/*
	 * If we are assigning into an object value then call a user
	 * defined function to do that.
	 */
	if ((v1->v_type == V_ADDR) && (v1->v_addr->v_type == V_OBJ))
	{
		(void) ObjectCall(OBJ_ASSIGN, v1->v_addr,
			&dataValue, NULL_VALUE);

		qfree(dataValue.v_num);

		return;
	}

	/*
	 * Store the value into the address.
	 */
	StoreValueIntoAddress(v1, &dataValue);
}


/*
 * Possibly set a variable's value to null and push the null value
 * onto the stack.  A child node is optional for this node.
 * This is used to explicitly bypass any defined object assignment
 * function so that a variable's type can be changed.
 */
void
ocSetNull(const Node * node)
{
	VALUE *	v1;
	VALUE	dataValue;

	/*
	 * If there is no kid then just push a null value.
	 */
	if (node->count == 0)
	{
		(void) NewValue();

		return;
	}

	/*
	 * Calculate the address to be assigned.
	 */
	v1 = CalcAddress(node->kids[0]);

	/*
	 * Get a null value.
	 */
	dataValue.v_type = V_NULL;
	dataValue.v_subtype = V_NOSUBTYPE;

	/*
	 * Store the value into the address.
	 */
	StoreValueIntoAddress(v1, &dataValue);

	/*
	 * Replace the stack value with a null.
	 */
	PopValue();
	(void) NewValue();
}


/*
 * Link to the value of our kid and push the linked value onto the
 * evaluation stack.
 */
void
ocLinkValue(const Node * node)
{
	VALUE *	v1;
	VALUE	result;

	v1 = CalcValue(node->kids[0]);

	LinkValue(v1, &result);

	PopValue();
	PushValue(&result);
}


/*
 * Copy the value of our kid and push the copied value onto the
 * evaluation stack.  Some values require no copying so that the
 * stack value is good as it is.
 */
void
ocCopyValue(const Node * node)
{
	VALUE *	v1;
	VALUE	result;

	v1 = CalcRawValue(node->kids[0]);

	switch (v1->v_type)
	{
		case V_NUM:
		case V_NULL:
		case V_STR:
		case V_COM:
			return;

		case V_ADDR:
			copyvalue(v1->v_addr, &result);

			break;

		case V_SUBADDR:
			result = LoadValueFromAddress(v1);

			break;

		default:
			if (!IsValueCopyNeeded(v1))
				return;

			copyvalue(v1, &result);

			break;
	}

	/*
	 * A new result was calculated.
	 * Replace the stack value with it.
	 */
	PopValue();
	PushValue(&result);
}


/*
 * Add a value to a variable and push the variable's address onto the
 * evaluation stack.
 */
void
ocPlusEquals(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	oldValue;
	VALUE	newValue;

	v1 = CalcAddress(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Try to add into the value without replacing it.
	 */
	if (ThisAddValue(v1, v2))
	{
		PopValue();

		return;
	}

	/*
	 * That failed, so we will calculate a totally new value
	 * and replace the varible with the new value.
	 */
	oldValue = LoadValueFromAddress(v1);

	addvalue(&oldValue, v2, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);

	PopValue();
}


/*
 * Subtract a value from a variable and push the variable's address onto
 * the evaluation stack.
 */
void
ocMinusEquals(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	oldValue;
	VALUE	newValue;

	v1 = CalcAddress(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Try to subtract from the value without replacing it.
	 */
	if (ThisSubtractValue(v1, v2))
	{
		PopValue();

		return;
	}

	/*
	 * That failed, so we will calculate a totally new value
	 * and replace the varible with the new value.
	 */
	oldValue = LoadValueFromAddress(v1);

	subvalue(&oldValue, v2, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);

	PopValue();
}


/*
 * Multiply a variable by a value and push the variable's address onto
 * the evaluation stack.
 */
void
ocMultiplyEquals(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	oldValue;
	VALUE	newValue;

	v1 = CalcAddress(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Try to multiply the value without replacing it.
	 */
	if (ThisMultiplyValue(v1, v2))
	{
		PopValue();

		return;
	}

	/*
	 * That failed, so we will calculate a totally new value
	 * and replace the varible with the new value.
	 */
	oldValue = LoadValueFromAddress(v1);

	mulvalue(&oldValue, v2, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);

	PopValue();
}


/*
 * Divide a variable by a value and push the variable's address onto
 * the evaluation stack.
 */
void
ocDivideEquals(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	oldValue;
	VALUE	newValue;

	v1 = CalcAddress(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Try to divide the value without replacing it.
	 */
	if (ThisDivideValue(v1, v2))
	{
		PopValue();

		return;
	}

	/*
	 * That failed, so we will calculate a totally new value
	 * and replace the varible with the new value.
	 */
	oldValue = LoadValueFromAddress(v1);

	divvalue(&oldValue, v2, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);

	PopValue();
}


/*
 * Integer divide a variable by a value and push the variable's address
 * onto the evaluation stack.
 */
void
ocQuotientEquals(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	oldValue;
	VALUE	newValue;
	VALUE	ignoreValue;

	v1 = CalcAddress(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Try to quotient the value without replacing it.
	 */
	if (ThisQuotientValue(v1, v2))
	{
		PopValue();

		return;
	}

	/*
	 * That failed, so we will calculate a totally new value
	 * and replace the varible with the new value.
	 */
	oldValue = LoadValueFromAddress(v1);

	ignoreValue.v_type = V_NULL;

	quovalue(&oldValue, v2, &ignoreValue, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);

	PopValue();
}


/*
 * Take the modulo of a variable and push the variable's address onto
 * the evaluation stack.
 */
void
ocModuloEquals(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	oldValue;
	VALUE	newValue;
	VALUE	ignoreValue;

	v1 = CalcAddress(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Try to modulo the value without replacing it.
	 */
	if (ThisModuloValue(v1, v2))
	{
		PopValue();

		return;
	}

	/*
	 * That failed, so we will calculate a totally new value
	 * and replace the varible with the new value.
	 */
	oldValue = LoadValueFromAddress(v1);

	ignoreValue.v_type = V_NULL;

	modvalue(&oldValue, v2, &ignoreValue, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);

	PopValue();
}


/*
 * AND a variable with a value and push the variable's address onto
 * the evaluation stack.
 */
void
ocAndEquals(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	oldValue;
	VALUE	newValue;

	v1 = CalcAddress(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Try to AND the value without replacing it.
	 */
	if (ThisAndValue(v1, v2))
	{
		PopValue();

		return;
	}

	/*
	 * That failed, so we will calculate a totally new value
	 * and replace the varible with the new value.
	 */
	oldValue = LoadValueFromAddress(v1);

	andvalue(&oldValue, v2, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);

	PopValue();
}


/*
 * OR a variable with a value and push the variable's address onto
 * the evaluation stack.
 */
void
ocOrEquals(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	oldValue;
	VALUE	newValue;

	v1 = CalcAddress(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Try to OR the value without replacing it.
	 */
	if (ThisOrValue(v1, v2))
	{
		PopValue();

		return;
	}

	/*
	 * That failed, so we will calculate a totally new value
	 * and replace the varible with the new value.
	 */
	oldValue = LoadValueFromAddress(v1);

	orvalue(&oldValue, v2, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);

	PopValue();
}


/*
 * Left shift a variable by a value and push the variable's address
 * onto the evaluation stack.
 */
void
ocLeftShiftEquals(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	oldValue;
	VALUE	newValue;

	v1 = CalcAddress(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Try to left shift the value without replacing it.
	 */
	if (ThisLeftShiftValue(v1, v2))
	{
		PopValue();

		return;
	}

	/*
	 * That failed, so we will calculate a totally new value
	 * and replace the varible with the new value.
	 */
	oldValue = LoadValueFromAddress(v1);

	shiftvalue(&oldValue, v2, FALSE, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);

	PopValue();
}


/*
 * Right shift a variable by a value and push the variable's address
 * onto the evaluation stack.
 */
void
ocRightShiftEquals(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	oldValue;
	VALUE	newValue;

	v1 = CalcAddress(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Try to right shift the value without replacing it.
	 */
	if (ThisRightShiftValue(v1, v2))
	{
		PopValue();

		return;
	}

	/*
	 * That failed, so we will calculate a totally new value
	 * and replace the varible with the new value.
	 */
	oldValue = LoadValueFromAddress(v1);

	shiftvalue(&oldValue, v2, TRUE, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);

	PopValue();
}


/*
 * Raise a variable to the power of a value and push the variable's
 * address onto the evaluation stack.
 */
void
ocPowerEquals(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	oldValue;
	VALUE	newValue;

	v1 = CalcAddress(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Try to raise the value to a power without replacing it.
	 */
	if (ThisPowerValue(v1, v2))
	{
		PopValue();

		return;
	}

	/*
	 * That failed, so we will calculate a totally new value
	 * and replace the varible with the new value.
	 */
	oldValue = LoadValueFromAddress(v1);

	powivalue(&oldValue, v2, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);

	PopValue();
}


/*
 * Swap the values of two variables.
 * Currently this does not allow the swapping of sub-values.
 * A null result is pushed onto the stack.
 */
void
ocSwap(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	tempValue;

	v1 = CalcFullAddress(node->kids[0]);
	v2 = CalcFullAddress(node->kids[1]);

	tempValue = v1->v_addr[0];
	v1->v_addr[0] = v2->v_addr[0];
	v2->v_addr[0] = tempValue;

	PopValue();
	PopValue();

	(void) NewValue();
}


/*
 * Evaluate the list of expressions and push the result of the last one.
 * This implements a comma separated expression list.
 */
void
ocExpressionList(const Node * node)
{
	int	index;

	/*
	 * Iterate over all of our kids.
	 */
	for (index = 0; index < node->count; index++)
	{
		if (index != 0)
			PopValue();

		CalcRawValue(node->kids[index]);
	}
}


/*
 * Calculate and push the result of an alternate (question mark operator).
 * The kids are the value to be tested, the value to be calculated if the
 * tested value is true, and the value to be calculated if the tested value
 * is false.
 */
void
ocAlternate(const Node * node)
{
	VALUE *	v1;
	Node *	evalNode;

	/*
	 * Calculate the value of the expression to be checked.
	 */
	v1 = CalcValue(node->kids[0]);

	/*
	 * Select the appropriate kid according to whether the value
	 * is true or not.
	 */
	if (testvalue(v1))
		evalNode = node->kids[1];
	else
		evalNode = node->kids[2];

	/*
	 * Remove the test value and replace it with the appropriate
	 * kid's value.
	 */
	PopValue();

	CalcValue(evalNode);
}


/*
 * Calculate a logical OR expression.
 * The kids are the expressions to be evaluated in turn until the
 * first one of them returns a value which is "true".  The true value
 * is pushed onto the evaluation stack.  If none of them are true
 * then a zero value is pushed onto the evaluation stack.
 */
void
ocLogicalOr(const Node * node)
{
	VALUE *	v1;
	int	index;

	/*
	 * Evaluate all of the kid nodes in order.
	 */
	for (index = 0; index < node->count; index++)
	{
		/*
		 * Calculate this kid's value and return it if it is true.
		 */
		v1 = CalcValue(node->kids[index]);

		if (testvalue(v1))
			return;

		/*
		 * Pop off the kid's value and check the next one.
		 */
		PopValue();
	}

	/*
	 * None of the kids were true, so push a zero value.
	 */
	PushZero();
}


/*
 * Calculate a logical AND expression.
 * The kids are the expressions to be evaluated in turn until the
 * first one of them returns a value which is "false".  The false value
 * is pushed onto the evaluation stack.  If none of them are false
 * then a true value is pushed onto the evaluation stack.
 */
void
ocLogicalAnd(const Node * node)
{
	VALUE *	v1;
	int	index;

	/*
	 * Evaluate all of the kid nodes in order.
	 */
	for (index = 0; index < node->count; index++)
	{
		/*
		 * Calculate this kid's value and return it if it is false.
		 */
		v1 = CalcValue(node->kids[index]);

		if (!testvalue(v1))
			return;

		/*
		 * Pop off the kid's value and check the next one.
		 */
		PopValue();
	}

	/*
	 * None of the kids were false, so push a one value.
	 */
	PushOne();
}


/*
 * Push a string literal value onto the evaluation stack.
 */
void
ocString(const Node * node)
{
	VALUE *	result;

	result = NewValue();

	result->v_str = slink(findstring(node->id));
	result->v_subtype = V_NOSUBTYPE;
	result->v_type = V_STR;
}


/*
 * Push the null value onto the evaluation stack.
 */
void
ocNull(const Node * node)
{
	(void) NewValue();
}


/*
 * Push a small integer onto the evaluation stack.
 */
void
ocIntLiteral(const Node * node)
{
	NUMBER *q;

	q = itoq(node->id);

	PushNumber(q);
}


/*
 * Push a number constant onto the evaluation stack.
 */
void
ocNumber(const Node * node)
{
	NUMBER *q;

	q = constvalue(node->id);

	if (q == NULL)
	{
		math_error("Numeric constant value not found");
		/*NOTREACHED*/
	}

	PushNumber(qlink(q));
}


/*
 * Add the values of all of our kids and push that result onto
 * the evaluation stack.
 */
void
ocAdd(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	int	index;
	VALUE	tempValue;

	/*
	 * Calculate the first kid's value.
	 */
	v1 = CalcValue(node->kids[0]);

	/*
	 * Add all of the other kid's values to it.
	 */
	for (index = 1; index < node->count; index++)
	{
		v2 = CalcValue(node->kids[index]);

		/*
		 * Add the top two stack values together to
		 * form a single stack value.
		 */
		addvalue(v1, v2, &tempValue);
		PopValue();
		PopValue();
		PushValue(&tempValue);
		v1 = TopValue();
	}
}


/*
 * Subtract the values of its two kids and push the result
 * onto the evaluation stack.
 */
void
ocSubtract(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	tempValue;

	/*
	 * Calculate the two kid's values.
	 */
	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Subtract the top two stack values together to
	 * form a single stack value.
	 */
	subvalue(v1, v2, &tempValue);
	PopValue();
	PopValue();
	PushValue(&tempValue);
}


/*
 * Multiply the values of all of the kids and push the result
 * onto the evaluation stack.
 */
void
ocMultiply(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	int	index;
	VALUE	tempValue;

	/*
	 * Calculate the first kid's value.
	 */
	v1 = CalcValue(node->kids[0]);

	/*
	 * Multiply all of the other kid's values to it.
	 */
	for (index = 1; index < node->count; index++)
	{
		v2 = CalcValue(node->kids[index]);

		/*
		 * Multiply the top two stack values together to
		 * form a single stack value.
		 */
		mulvalue(v1, v2, &tempValue);
		PopValue();
		PopValue();
		PushValue(&tempValue);
		v1 = TopValue();
	}
}


/*
 * Raise the first kid to the power of the second kid and
 * push the result onto the evaluation stack.
 */
void
ocPower(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	tempValue;

	/*
	 * Calculate the two kid's values.
	 */
	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Use the top two stack values together to
	 * form a single stack value.
	 */
	powivalue(v1, v2, &tempValue);
	PopValue();
	PopValue();
	PushValue(&tempValue);
}


/*
 * Calculate the factorial of our kid and push the result onto
 * the evaluation stack.
 */
void
ocFactorial(const Node * node)
{
	VALUE *	v1;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);

	switch (v1->v_type)
	{
		case V_OBJ:
			tempValue = ObjectCall(OBJ_FACT, v1,
				NULL_VALUE, NULL_VALUE);

			break;

		case V_NUM:
			tempValue.v_num = qfact(v1->v_num);
			tempValue.v_type = V_NUM;
			tempValue.v_subtype = V_NOSUBTYPE;

			break;

		default:
			math_error("Non-real argument for factorial");
			/*NOTREACHED*/
	}

	PopValue();
	PushValue(&tempValue);
}


/*
 * Divide the first kid by the second kid and push the result
 * onto the evaluation stack.
 */
void
ocDivide(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	tempValue;

	/*
	 * Calculate the two kid's values.
	 */
	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Divide the top two stack values together to
	 * form a single stack value.
	 */
	divvalue(v1, v2, &tempValue);
	PopValue();
	PopValue();
	PushValue(&tempValue);
}


/*
 * Take the quotient of the first kid by the second kid and
 * push the result onto the evaluation stack.
 */
void
ocQuotient(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	tempValue;
	VALUE	ignoreValue;

	/*
	 * Calculate the two kid's values.
	 */
	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Divide the top two stack values together to
	 * form a single stack value.
	 */
	ignoreValue.v_type = V_NULL;

	quovalue(v1, v2, &ignoreValue, &tempValue);

	PopValue();
	PopValue();
	PushValue(&tempValue);
}


/*
 * Reduce the first kid by the second kid's modulus and push
 * the result onto the evaluation stack.
 */
void
ocModulo(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	tempValue;
	VALUE	ignoreValue;

	/*
	 * Calculate the two kid's values.
	 */
	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * Divide the top two stack values together to
	 * form a single stack value.
	 */
	ignoreValue.v_type = V_NULL;

	modvalue(v1, v2, &ignoreValue, &tempValue);

	PopValue();
	PopValue();
	PushValue(&tempValue);
}


/*
 * Produce a quotient and a remainder from the result of a division
 * and assign those values to two variables, and push zero or one onto
 * the evaluation stack according to whether the division is exact.
 */
void
ocQuotientMod(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE *	v3;
	VALUE *	v4;
	VALUE	valQuo;
	VALUE	valMod;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);
	v3 = CalcFullAddress(node->kids[2]);
	v4 = CalcFullAddress(node->kids[3]);

	if ((v1->v_type != V_NUM) || (v2->v_type != V_NUM))
	{
		math_error("Non-reals for quomod");
		/*NOTREACHED*/
	}

	/*
	 * Get the actual variables to be stored into.
	 */
	v3 = v3->v_addr;
	v4 = v4->v_addr;

	/*
	 * Perform the division to get the results.
	 */
	valQuo.v_type = V_NUM;
	valMod.v_type = V_NUM;

	result = qquomod(v1->v_num, v2->v_num, &valQuo.v_num, &valMod.v_num);

	/*
	 * Store the results back into the variables.
	 */
	freevalue(v3);
	freevalue(v4);
	*v3 = valQuo;
	*v4 = valMod;

	/*
	 * Replace the stack arguments with the boolean result
	 * indicating whether the result was exact.
	 */
	PopValue();
	PopValue();
	PopValue();
	PopValue();
	PushBoolean(result);
}


/*
 * Perform an AND on the two kids and push the result onto the
 * evaluation stack.
 */
void
ocAnd(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	NUMBER *q;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	if ((v1->v_type != V_NUM) || (v2->v_type != V_NUM))
	{
		math_error("Non-numerics for and");
		/*NOTREACHED*/
	}

	q = qand(v1->v_num, v2->v_num);

	PopValue();
	PopValue();
	PushNumber(q);
}


/*
 * Perform an OR on the two kids and push the result on the evaluation stack.
 */
void
ocOr(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	NUMBER *q;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	if ((v1->v_type != V_NUM) || (v2->v_type != V_NUM))
	{
		math_error("Non-numerics for or");
		/*NOTREACHED*/
	}

	q = qor(v1->v_num, v2->v_num);

	PopValue();
	PopValue();
	PushNumber(q);
}


/*
 * Perform a logical NOT on our kid and push the result onto
 * the evaluation stack.
 */
void
ocNot(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = !testvalue(v1);

	PopValue();
	PushBoolean(result);
}


/*
 * Negate the value of our kid and push the result onto the
 * evaluation stack.
 */
void
ocNegate(const Node * node)
{
	VALUE *	v1;
	NUMBER *q;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);

	/*
	 * If the value is a number then do a short cut.
	 */
	if (v1->v_type == V_NUM)
	{
		q = qneg(v1->v_num);

		PopValue();
		PushNumber(q);

		return;
	}

	/*
	 * Do the general negation.
	 */
	negvalue(v1, &tempValue);
	PopValue();
	PushValue(&tempValue);
}


/*
 * Invert the value of our kid and push the result onto the
 * evaluation stack.
 */
void
ocInvert(const Node * node)
{
	VALUE *	v1;
	NUMBER *q;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);

	/*
	 * If the value is a number then short cut the inversion.
	 */
	if (v1->v_type == V_NUM)
	{
		q = qinv(v1->v_num);

		PopValue();
		PushNumber(q);

		return;
	}

	/*
	 * Do the general inversion.
	 */
	invertvalue(v1, &tempValue);
	PopValue();
	PushValue(&tempValue);
}


/*
 * Perform a scaling operation based on our two kids and push the
 * result onto the evaluation stack.
 */
void
ocScale(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	scalevalue(v1, v2, &tempValue);

	PopValue();
	PopValue();
	PushValue(&tempValue);
}


/*
 * Take the integer part of our kid's value and push the result
 * onto the evaluation stack.
 */
void
ocIntegerPart(const Node * node)
{
	VALUE *	v1;
	NUMBER *q;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);

	/*
	 * If the value is a number then short cut the operation.
	 */
	if (v1->v_type == V_NUM)
	{
		q = qint(v1->v_num);

		PopValue();
		PushNumber(q);

		return;
	}

	/*
	 * Do the general integer value.
	 */
	intvalue(v1, &tempValue);
	PopValue();
	PushValue(&tempValue);
}


/*
 * Take the fractional part of our kid's value and push the
 * result onto the evaluation stack.
 */
void
ocFractionalPart(const Node * node)
{
	VALUE *	v1;
	NUMBER *q;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);

	/*
	 * If the value is a number then short cut the operation.
	 */
	if (v1->v_type == V_NUM)
	{
		q = qfrac(v1->v_num);

		PopValue();
		PushNumber(q);

		return;
	}

	/*
	 * Do the general fractional value.
	 */
	fracvalue(v1, &tempValue);
	PopValue();
	PushValue(&tempValue);
}


/*
 * Take the absolute value of our kid's value and push the result
 * onto the evaluation stack.  The second kid is optional and gives
 * the epsilon value for values which aren't exact.
 */
void
ocAbsoluteValue(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	NUMBER *q;
	VALUE	tempValue;
	VALUE	epsilonValue;

	v1 = CalcValue(node->kids[0]);

	/*
	 * If a second kid was specified then use it as the epsilon value
	 * otherwise use the default epsilon value.
	 */
	if (node->count > 1)
	{
		v2 = CalcValue(node->kids[1]);
	}
	else
	{
		v2 = &epsilonValue;

		v2->v_type = V_NUM;
		v2->v_subtype = V_NOSUBTYPE;
		v2->v_num = conf->epsilon;
	}

	/*
	 * If the kids are not normal numbers then do the
	 * generic case.
	 */
	if ((v1->v_type != V_NUM) || (v2->v_type != V_NUM) ||
		!qispos(v2->v_num))
	{
		absvalue(v1, v2, &tempValue);

		PopValue();

		if (node->count > 1)
			PopValue();

		PushValue(&tempValue);

		return;
	}

	/*
	 * This is a normal number.
	 */
	q = qqabs(v1->v_num);

	PopValue();

	if (node->count > 1)
		PopValue();

	PushNumber(q);
}


/*
 * Calculate the norm of our kid's value and push the result
 * onto the evaluation stack.
 */
void
ocNorm(const Node * node)
{
	VALUE *	v1;
	NUMBER *q;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);

	/*
	 * If this is a normal number then short cut the operation.
	 */
	if (v1->v_type == V_NUM)
	{
		q = qsquare(v1->v_num);

		PopValue();
		PushNumber(q);

		return;
	}

	/*
	 * Do the general norm.
	 */
	normvalue(v1, &tempValue);

	PopValue();
	PushValue(&tempValue);
}


/*
 * Square the value of our kid and push the result onto the
 * evaluation stack.
 */
void
ocSquare(const Node * node)
{
	VALUE *	v1;
	NUMBER *q;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);

	/*
	 * If the value is a number then short cut the squaring.
	 */
	if (v1->v_type == V_NUM)
	{
		q = qsquare(v1->v_num);

		PopValue();
		PushNumber(q);

		return;
	}

	/*
	 * Do the general square.
	 */
	squarevalue(v1, &tempValue);

	PopValue();
	PushValue(&tempValue);
}


/*
 * Push the length of a string onto the evaluation stack.
 */
void
ocStrlen(const Node * node)
{
	VALUE *		v1;
	NUMBER *	q;

	v1 = CalcValue(node->kids[0]);

	if (v1->v_type != V_STR)
		math_error("Strlen of non-string");

	q = itoq(v1->v_str->s_len);

	PopValue();
	PushNumber(q);
}


/*
 * Push one or zero onto the evaluation stack according to whether the
 * values of our two kids are of the same type.
 */
void
ocIsType(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	if ((v1->v_type != V_OBJ) || (v2->v_type != V_OBJ))
		result = (v1->v_type == v2->v_type);
	else
		result = (v1->v_obj->o_actions == v2->v_obj->o_actions);

	PopValue();
	PopValue();
	PushBoolean(result);
}


/*
 * Push a zero or one value onto the evaluation stack according to
 * whether our kid's value is a normal integer.
 */
void
ocIsInteger(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	/*
	 * If the value is not a number then push a zero value.
	 */
	if (v1->v_type != V_NUM)
	{
		PopValue();
		PushZero();

		return;
	}

	/*
	 * The value is a number, so push a zero or a one value
	 * according to whether it is an integer.
	 */
	result = qisint(v1->v_num);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a zero or one value onto the evaluation stack according to
 * whether our kid's value is a number (real or complex).
 */
void
ocIsNumber(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = ((v1->v_type == V_NUM) || (v1->v_type == V_COM));

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is a matrix.
 */
void
ocIsMatrix(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_LIST);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is a list.
 */
void
ocIsList(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_LIST);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is an error.
 */
void
ocIsError(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_ERROR);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is a function reference.
 */
void
ocIsFunction(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_FUNC);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is a memory block.
 */
void
ocIsBlock(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = ((v1->v_type == V_BLOCK) || (v1->v_type == V_NBLOCK));

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is an object.
 */
void
ocIsObject(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_OBJ);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's string value names a defined type of object.
 */
void
ocIsObjectType(const Node * node)
{ 
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	if (v1->v_type != V_STR)
	{
		math_error("Non-string argument for isobjtype");
		/*NOTREACHED*/
	}

	result = (checkobject(v1->v_str->s_str) >= 0);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is a string.
 */
void
ocIsString(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_STR);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is a file handle.
 */
void
ocIsFile(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_FILE);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is a rand number generator.
 */
void
ocIsRand(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_RAND);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is a random number generator.
 */
void
ocIsRandom(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_RANDOM);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is a configuration value.
 */
void
ocIsConfig(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_CONFIG);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is a hash value.
 */
void
ocIsHash(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_HASH);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a one or zero value onto the evaluation stack according to
 * whether our kid's value is an association.
 */
void
ocIsAssociation(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_ASSOC);

	PopValue();
	PushBoolean(result);
}


/*
 * Push one or zero onto the evaluation stack according to whether
 * our kid's value is simple.
 */
void
ocIsSimple(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	switch (v1->v_type)
	{
		case V_NULL:
		case V_NUM:
		case V_COM:
		case V_STR:
			result = TRUE;
			break;

		default:
			result = FALSE;
			break;
	}

	PopValue();
	PushBoolean(result);
}


/*
 * Push zero or one onto the evaluation stack according to whether
 * our kid's value is an even or odd integer.
 */
void
ocIsOdd(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = ((v1->v_type == V_NUM) && qisodd(v1->v_num));

	PopValue();
	PushBoolean(result);
}


/*
 * Push zero or one onto the evaluation stack according to whether
 * our kid's value is an odd or even integer.
 */
void
ocIsEven(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = ((v1->v_type == V_NUM) && qiseven(v1->v_num));

	PopValue();
	PushBoolean(result);
}


/*
 * Push zero or one onto the evaluation stack according to whether
 * our kid's value is a real number.
 */
void
ocIsReal(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_NUM);

	PopValue();
	PushBoolean(result);
}


/*
 * Push one or zero onto the evaluation stack according to whether
 * our kid's value is null.
 */
void
ocIsNull(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = (v1->v_type == V_NULL);

	PopValue();
	PushBoolean(result);
}


/*
 * Push a zero or one value onto the evaluation stack according to
 * whether the bit specified by the second kid is set in the value of
 * the first kid.
 */
void
ocBit(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	long	index;
	int	r;
	VALUE	result;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * The bit position has to be an integer.
	 */
	if ((v2->v_type != V_NUM) || qisfrac(v2->v_num))
	{
		PopValue();
		PopValue();

		result = error_value(E_BIT1);
		PushValue(&result);

		return;
	}

	/*
	 * The bit position can't be too large.
	 */
	if (zge31b(v2->v_num->num))
	{
		PopValue();
		PopValue();

		result = error_value(E_BIT2);
		PushValue(&result);

		return;
	}

	/*
	 * Get the bit number.
	 */
	index = qtoi(v2->v_num);

	/*
	 * Switch on the type of value to get the bit.
	 */
	switch (v1->v_type)
	{
		case V_NUM:
			r = qisset(v1->v_num, index);
			break;

		case V_STR:
			r = stringbit(v1->v_str, index);
			break;

		default:
			r = 2;
	}

	PopValue();
	PopValue();

	result.v_subtype = V_NOSUBTYPE;

	if (r > 1)
		result = error_value(E_BIT1);
	else if (r < 0)
		result.v_type = V_NULL;
	else
	{
		result.v_type = V_NUM;
		result.v_num = itoq(r);
	}

	PushValue(&result);
}


/*
 * Push the bit number of the highest set bit in a value onto the
 * evaluation stack.
 */
void
ocHighBit(const Node * node)
{
	VALUE *	v1;
	long	index;
	VALUE	result;

	v1 = CalcValue(node->kids[0]);

	switch (v1->v_type)
	{
		case V_NUM:
			if (qiszero(v1->v_num))
			{
				index = -1;
				break;
			}

			if (qisfrac(v1->v_num))
			{
				index = -2;
				break;
			}

			index = zhighbit(v1->v_num->num);

			break;

		case V_STR:
			index = stringhighbit(v1->v_str);
			break;

		default:
			index = -3;
	}

	PopValue();

	switch (index)
	{
		case -3:
			result = error_value(E_HIGHBIT1);
			break;

		case -2:
			result = error_value(E_HIGHBIT2);
			break;

		default:
			PushNumber(itoq(index));

			return;
	}

	PushValue(&result);
}


/*
 * Push the bit number of the lowest set bit in a value onto the
 * evaluation stack.
 */
void
ocLowBit(const Node * node)
{
	VALUE *	v1;
	long	index;
	VALUE	result;

	v1 = CalcValue(node->kids[0]);

	switch (v1->v_type)
	{
		case V_NUM:
			if (qiszero(v1->v_num))
			{
				index = -1;
				break;
			}

			if (qisfrac(v1->v_num))
			{
				index = -2;
				break;
			}

			index = zlowbit(v1->v_num->num);

			break;

		case V_STR:
			index = stringlowbit(v1->v_str);
			break;

		default:
			index = -3;
	}

	PopValue();

	switch (index)
	{
		case -3:
			result = error_value(E_LOWBIT1);
			break;

		case -2:
			result = error_value(E_LOWBIT2);
			break;

		default:
			PushNumber(itoq(index));

			return;
	}

	PushValue(&result);
}


/*
 * Push the real part of our kid's value onto the evaluation stack.
 */
void
ocRealPart(const Node * node)
{
	VALUE *	v1;
	NUMBER *q;

	v1 = CalcValue(node->kids[0]);

	switch (v1->v_type)
	{
		case V_NUM:
			q = v1->v_num;
			break;

		case V_COM:
			q = v1->v_com->real;
			break;

		default:
			math_error("Taking real part of non-number");
			return;
	}

	q = qlink(q);

	PopValue();
	PushNumber(q);
}


/*
 * Push the imaginary part of our kid's value onto the evaluation stack.
 */
void
ocImaginaryPart(const Node * node)
{
	VALUE *	v1;
	NUMBER *q;

	v1 = CalcValue(node->kids[0]);

	switch (v1->v_type)
	{
		case V_NUM:
			q = &_qzero_;
			break;

		case V_COM:
			q = v1->v_com->imag;
			break;

		default:
			math_error("Taking imaginary part of non-number");
			return;
	}

	q = qlink(q);

	PopValue();
	PushNumber(q);
}


/*
 * Take the complex conjugate of our kid's value.
 */
void
ocConjugate(const Node * node)
{
	VALUE *	v1;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);

	conjvalue(v1, &tempValue);

	PopValue();
	PushValue(&tempValue);
}


/*
 * Calculate the "fast index" address for a matrix or other data type.
 * The kids are the object to be indexed and the index value.
 * The calculated address is pushed onto the evaluation stack.
 */
void
ocFastIndexAddr(const Node * node)
{
	VALUE *		v1;
	VALUE *		v2;
	VALUE *		result;
	VALUE *		address;
	NUMBER *	q;
	MATRIX *	mat;
	LIST *		list;
	ASSOC *		assoc;
	long		index;

	/*
	 * Evaluate the object and the index value.
	 * TODO: Allow sub-addresses.
	 */
	v1 = CalcFullAddress(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	/*
	 * The index value must be a non-negative integer.
	 */
	if (v2->v_type != V_NUM)
	{
		math_error("Fast indexing by non-number");
		/*NOTREACHED*/
	}

	q = v2->v_num;

	if (qisfrac(q))
	{
		math_error("Fast indexing by non-integer");
		/*NOTREACHED*/
	}

	if (zge31b(q->num) || (qisneg(q) < 0))
	{
		math_error("Index out of range for fast indexing");
		/*NOTREACHED*/
	}

	index = qtoi(q);

	PopValue();

	/*
	 * Switch on the type to evaluate it into address.
	 * If address remains null then the index was out of bounds.
	 */
	address = 0;

	switch (v1->v_addr->v_type)
	{
		case V_OBJ:
			if (index < v1->v_addr->v_obj->o_actions->oa_count)
				address = v1->v_addr->v_obj->o_table + index;

			break;

		case V_MAT:
			mat = v1->v_addr->v_mat;

			if (index < mat->m_size)
				address = mat->m_table + index;

			break;

		case V_LIST:
			list = v1->v_addr->v_list;

			address = listfindex(list, index);

			break;

		case V_ASSOC:
			assoc = v1->v_addr->v_assoc;

			address = assocfindex(assoc, index);

			break;

		default:
			math_error("Bad variable type for fast indexing");
			/*NOTREACHED*/
	}

	PopValue();

	/*
	 * Check whether the address was calculated.
	 */
	if (address == 0)
	{
		math_error("Fast index value too large");
		/*NOTREACHED*/
	}

	/*
	 * Push the address onto the evaluation stack.
	 */
	result = NewValue();
	result->v_addr = address;
	result->v_type = V_ADDR;
}


/*
 * Push -1, 0, or 1 onto the evaluation stack according to whether
 * our kid's value is negative, zero, or positive.
 */
void
ocSgn(const Node * node)
{
	VALUE *	v1;
	NUMBER *q;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);

	if (v1->v_type == V_NUM)
	{
		q = qsign(v1->v_num);

		PopValue();
		PushNumber(q);

		return;
	}

	sgnvalue(v1, &tempValue);

	PopValue();
	PushValue(&tempValue);
}


/*
 * Push 0 or 1 onto the evaluation stack according to whether our
 * kid's value is "nonzero".
 */
void
ocTest(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = testvalue(v1);

	PopValue();
	PushBoolean(result);
}


/*
 * Push the numerator of our kid's value onto the evaluation stack.
 */
void
ocNumerator(const Node * node)
{
	VALUE *	v1;
	NUMBER *q;

	v1 = CalcValue(node->kids[0]);

	if (v1->v_type != V_NUM)
	{
		math_error("Numerator of non-number");
		/*NOTREACHED*/
	}

	q = qnum(v1->v_num);

	PopValue();
	PushNumber(q);
}


/*
 * Push the denominator of our kid's value on the evaluation stack.
 */
void
ocDenominator(const Node * node)
{
	VALUE *	v1;
	NUMBER *q;

	v1 = CalcValue(node->kids[0]);

	if (v1->v_type != V_NUM)
	{
		math_error("Denominator of non-number");
		/*NOTREACHED*/
	}

	q = qden(v1->v_num);

	PopValue();
	PushNumber(q);
}


/*
 * Call a user defined function and push the result onto the evaluation
 * stack.  Our kids are the arguments to the function.  The function id
 * is hard-coded into the node.
 */
void
ocUserCall(const Node * node)
{
	int		index;
	UserFunc *	func;
	VALUE *		args;
	VALUE		result;

	/*
	 * Find the function to be called.
	 */
	func = FindUserFunction(node->id);

	if (func == NULL)
	{
		math_error("Function \"%s\" is undefined",
			NameUserFunction(node->id));

		/*NOTREACHED*/
	}

	/*
	 * Evaluate all of the kids in order and push their values
	 * onto the evaluation stack.  Remember the address of the
	 * first argument so as to form a base of the arguments.
	 */
	args = 0;

	for (index = 0; index < node->count; index++)
	{
		CalcValue(node->kids[index]);

		if (args == 0)
			args = TopValue();
	}

	/*
	 * Call the function.
	 * Since this is a normal function, there is no THIS value.
	 */
	result = Calculate(func, 0, 0, args, node->count);

	/*
	 * Pop off all of the arguments.
	 */
	for (index = node->count; index > 0; index--)
		PopValue();

	/*
	 * Push the result back onto the evaluation stack.
	 */
	PushValue(&result);
}


/*
 * Perform a built-in function call and push the result onto the
 * evaluation stack.  Our kids are the arguments to the function.
 * The function id to call is hard-coded into the node.
 */
void
ocBuiltinCall(const Node * node)
{
	int	index;
	VALUE *	args;
	VALUE	result;

	/*
	 * Evaluate all of the kids in order and push their raw values
	 * onto the evaluation stack.  Remember the address of the
	 * first argument so as to form a base of the arguments.
	 */
	args = 0;

	for (index = 0; index < node->count; index++)
	{
		CalcRawValue(node->kids[index]);

		if (args == 0)
			args = TopValue();
	}

	/*
	 * Call the function.
	 */
	result = CallBuiltinFunction(node->id, args, node->count);

	/*
	 * Pop off all of the arguments.
	 */
	for (index = node->count; index > 0; index--)
		PopValue();

	/*
	 * Push the result back onto the evaluation stack.
	 */
	PushValue(&result);
}


/*
 * Call a method for a value and push the result onto the evaluation
 * stack.  The first kid is the value which the method operates on.
 * The remaining kids are the arguments to the method.  The method id
 * is hard-coded into the node.  The data field of the node is non-nll
 * if this is a SUPER call.  The code which executes the method is
 * dependent on the type of value which is being operated on.
 */
void
ocMethodCall(const Node * node)
{
	int	argCount;
	int	index;
	int	superDepth;
	VALUE *	value;
	VALUE *	args;
	VALUE	result;

	/*
	 * Evaluate all of the kids in order and push their values
	 * onto the evaluation stack.  Remember the address of the
	 * first argument so as to form a base of the arguments.
	 */
	args = 0;

	for (index = 0; index < node->count; index++)
	{
		CalcValue(node->kids[index]);

		if (args == 0)
			args = TopValue();
	}

	/*
	 * Get the value the method is operating on from the first kid.
	 */
	value = args++;
	argCount = node->count - 1;

	/*
	 * Dereference an address.
	 */
	if (value->v_type == V_ADDR)
		value = value->v_addr;

	/*
	 * Determine the depth of a SUPER call.
	 * This is 0 if we are not doing a SUPER call, or is one more
	 * than our current super depth if we are.
	 */
	superDepth = 0;

	if (node->data != 0)
		superDepth = GetSuperDepth() + 1;

	/*
	 * Perform the method call for this method id and value.
	 */
	result = MethodCall(node->id, superDepth, value, args, argCount);

	/*
	 * Pop off all but the first argument.
	 */
	for (index = node->count; index > 1; index--)
		PopValue();

	/*
	 * If the result is an address value pointing to the first
	 * value on the evaluation stack then the method returned
	 * 'this' as the result.  In this case we just return leaving
	 * that value on the stack.
	 */
	if ((result.v_type == V_ADDR) && (result.v_addr == value))
		return;

	/*
	 * A value other than 'this was returned.
	 * Pop the first argument off too.
	 */
	PopValue();

	/*
	 * Push the result back onto the evaluation stack.
	 */
	PushValue(&result);
}


/*
 * Compare the values of our two kids and push a value of -1, 0, or 1
 * onto the evaluation stack according to their relative values.
 */
void
ocCompare(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	relvalue(v1, v2, &tempValue);

	PopValue();
	PopValue();

	PushValue(&tempValue);
}


/*
 * Compare the values of our two kids and push a 0 or 1 value onto
 * the evaluation stack according to whether their values have the
 * same address.
 */
void
ocEqualAddress(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	int	result;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	result = EqualAddressValue(v1, v2);

	PopValue();
	PopValue();
	PushBoolean(result != 0);
}


/*
 * Compare the values of our two kids and push a 0 or 1 value onto
 * the evaluation stack according to whether they are equal.
 */
void
ocEqual(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	int	result;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	result = comparevalue(v1, v2);

	PopValue();
	PopValue();
	PushBoolean(result == 0);
}


/*
 * Compare the values of our two kids and push a 0 or 1 value onto
 * the evaluation stack according to whether they are not equal.
 */
void
ocNotEqual(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	int	result;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	result = comparevalue(v1, v2);

	PopValue();
	PopValue();
	PushBoolean(result != 0);
}


/*
 * Compare the values of our two kids and push a 0 or 1 value onto
 * the evaluation stack according to whether or not the first value
 * is less than or equal to the second value.
 */
void
ocLessThanOrEqual(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	BOOL	result;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	relvalue(v1, v2, &tempValue);

	result = ((tempValue.v_type == V_NUM) && !qispos(tempValue.v_num));

	PopValue();
	PopValue();
	PushBoolean(result);
}


/*
 * Compare the values of our two kids and push a 0 or 1 value onto
 * the evaluation stack according to whether or not the first value
 * is greater than or equal to the second value.
 */
void
ocGreaterThanOrEqual(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	BOOL	result;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	relvalue(v1, v2, &tempValue);

	result = ((tempValue.v_type == V_NUM) && !qisneg(tempValue.v_num));

	PopValue();
	PopValue();
	PushBoolean(result);
}


/*
 * Compare the values of our two kids and push a 0 or 1 value onto
 * the evaluation stack according to whether or not the first value
 * is less than the second value.
 */
void
ocLessThan(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	BOOL	result;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	relvalue(v1, v2, &tempValue);

	result = ((tempValue.v_type == V_NUM) && qisneg(tempValue.v_num));

	PopValue();
	PopValue();
	PushBoolean(result);
}


/*
 * Compare the values of our two kids and push a 0 or 1 value onto
 * the evaluation stack according to whether or not the first value
 * is greater than the second value.
 */
void
ocGreaterThan(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	BOOL	result;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	relvalue(v1, v2, &tempValue);

	result = ((tempValue.v_type == V_NUM) && qispos(tempValue.v_num));

	PopValue();
	PopValue();
	PushBoolean(result);
}


/*
 * Compare the value of our kid against zero and push the result onto
 * the evaluation stack.
 */
void
ocEqualZero(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = !testvalue(v1);

	PopValue();
	PushBoolean(result);
}


/*
 * Compare the value of our kid against zero and push the result onto
 * the evaluation stack.
 */
void
ocNotEqualZero(const Node * node)
{
	VALUE *	v1;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	result = testvalue(v1);

	PopValue();
	PushBoolean(result);
}


/*
 * Compare the value of our kid against zero and push the result onto
 * the evaluation stack.
 */
void
ocGreaterThanZero(const Node * node)
{
	VALUE *	v1;
	VALUE	zeroValue;
	VALUE	tempValue;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	zeroValue.v_type = V_NUM;
	zeroValue.v_num = &_qzero_;

	relvalue(v1, &zeroValue, &tempValue);

	result = ((tempValue.v_type == V_NUM) && qispos(tempValue.v_num));

	PopValue();
	PushBoolean(result);
}


/*
 * Compare the value of our kid against zero and push the result onto
 * the evaluation stack.
 */
void
ocGreaterThanOrEqualZero(const Node * node)
{
	VALUE *	v1;
	VALUE	zeroValue;
	VALUE	tempValue;
	BOOL	result;

	v1 = CalcValue(node->kids[0]);

	zeroValue.v_type = V_NUM;
	zeroValue.v_num = &_qzero_;

	relvalue(v1, &zeroValue, &tempValue);

	result = ((tempValue.v_type == V_NUM) && !qisneg(tempValue.v_num));

	PopValue();
	PushBoolean(result);
}


/*
 * Increment a variable's value specified by our kid and push the
 * new value onto the evaluation stack.  (Actually, the original
 * address is left on the stack, which will provide the new value
 * if it is used again.)
 */
void
ocPreIncrement(const Node * node)
{
	VALUE *	v1;
	VALUE	oldValue;
	VALUE	newValue;

	v1 = CalcAddress(node->kids[0]);

	oldValue = LoadValueFromAddress(v1);

	incvalue(&oldValue, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);
}


/*
 * Decrement a variable's value specified by our kid and push the
 * new value onto the evaluation stack.  (Actually, the original
 * address is left on the stack, which will provide the new value
 * is used again.)
 */
void
ocPreDecrement(const Node * node)
{
	VALUE *	v1;
	VALUE	oldValue;
	VALUE	newValue;

	v1 = CalcAddress(node->kids[0]);

	oldValue = LoadValueFromAddress(v1);

	decvalue(&oldValue, &newValue);

	if (v1->v_type == V_SUBADDR)
		freevalue(&oldValue);

	StoreValueIntoAddress(v1, &newValue);
}


/*
 * Increment a variable's value specified by our kid and push the
 * old value onto the evaluation stack.
 */
void
ocPostIncrement(const Node * node)
{
	VALUE *	v1;
	VALUE	oldValue;
	VALUE	newValue;

	/*
	 * Get the address of the value to be incremented.
	 */
	v1 = CalcAddress(node->kids[0]);

	/*
	 * Get the old value from the address.
	 */
	oldValue = LoadValueFromAddress(v1);

	/*
	 * Increment the value.
	 */
	incvalue(&oldValue, &newValue);

	/*
	 * If this is a full address then clear out the destination
	 * value type so that it isn't freed.  We want to do this
	 * since we want that old value to push onto the stack.
	 */
	if (v1->v_type == V_ADDR)
	{
		v1->v_addr->v_type = V_NULL;
		v1->v_addr->v_subtype = V_NOSUBTYPE;
	}

	/*
	 * Store the new value into the address.
	 */
	StoreValueIntoAddress(v1, &newValue);

	/*
	 * Push the old value onto the stack.
	 */
	PopValue();
	PushValue(&oldValue);
}


/*
 * Decrement a variable's value specified by our kid and push the
 * old value onto the evaluation stack.
 */
void
ocPostDecrement(const Node * node)
{
	VALUE *	v1;
	VALUE	oldValue;
	VALUE	newValue;

	/*
	 * Get the address of the value to be decremented.
	 */
	v1 = CalcAddress(node->kids[0]);

	/*
	 * Get the old value from the address.
	 */
	oldValue = LoadValueFromAddress(v1);

	/*
	 * Decrement the value.
	 */
	decvalue(&oldValue, &newValue);

	/*
	 * If this is a full address then clear out the destination
	 * value type so that it isn't freed.  We want to do this
	 * since we want that old value to push onto the stack.
	 */
	if (v1->v_type == V_ADDR)
	{
		v1->v_addr->v_type = V_NULL;
		v1->v_addr->v_subtype = V_NOSUBTYPE;
	}

	/*
	 * Store the new value into the address.
	 */
	StoreValueIntoAddress(v1, &newValue);

	/*
	 * Push the old value onto the stack.
	 */
	PopValue();
	PushValue(&oldValue);
}


/*
 * Left shift one value by another one as specified by our kids and
 * push the result into the evaluation stack.
 */
void
ocLeftShift(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	shiftvalue(v1, v2, FALSE, &tempValue);

	PopValue();
	PopValue();
	PushValue(&tempValue);
}


/*
 * Right shift one value by another one as specified by our kids and
 * push the result into the evaluation stack.
 */
void
ocRightShift(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);
	v2 = CalcValue(node->kids[1]);

	shiftvalue(v1, v2, TRUE, &tempValue);

	PopValue();
	PopValue();
	PushValue(&tempValue);
}


/*
 * Push a zero value onto the evaluation stack.
 */
void
ocZero(const Node * node)
{
	PushZero();
}


/*
 * Push a one value onto the evaluation stack.
 */
void
ocOne(const Node * node)
{
	PushOne();
}


/*
 * Push a -1 value onto the evaluation stack.
 */
void
ocMinusOne(const Node * node)
{
	PushMinusOne();
}


/*
 * Push the "old" value onto the evaluation stack.
 * This is the result of the last top level expression.
 */
void
ocOldValue(const Node * node)
{
	VALUE *	result;
	VALUE	tempValue;

	GetOldValue(&tempValue);

	result = NewValue();

	copyvalue(&tempValue, result);
}


/*
 * Push a history value onto the top of the evaluation stack.
 * This is the result of some previous top level expression.
 * The selected value is determined by the id value of the node.
 */
void
ocHistoryValue(const Node * node)
{
	VALUE *	result;
	VALUE	tempValue;

	if (!GetHistoryValue(node->id, &tempValue))
		math_error("Out of range history value %ld", node->id);

	result = NewValue();

	copyvalue(&tempValue, result);
}


/*
 * Set or get the new epsilon value and push its old value onto the
 * evaluation stack.  Which we do depends on the number of kids.
 */
void
ocEpsilon(const Node * node)
{
	VALUE *		v1;
	NUMBER *	newEpsilon;
	NUMBER *	oldEpsilon;

	/*
	 * If there is no kid then return the current value.
	 */
	if (node->count == 0)
	{
		PushNumber(qlink(conf->epsilon));

		return;
	}

	/*
	 * We are setting a new value.
	 */
	v1 = CalcValue(node->kids[0]);

	if (v1->v_type != V_NUM)
		math_error("Non-number for setepsilon");

	newEpsilon = v1->v_num;

	if (!qispos(newEpsilon))
		math_error("Non-positive value for setepsilon");

	oldEpsilon = qlink(conf->epsilon);

	conf->epsilon = qlink(newEpsilon);

	PopValue();
	PushNumber(oldEpsilon);
}


/*
 * Set or get a configuration value according to the name and value kid
 * values.  Which we do is dependent on whether the value kid exists.
 * The old value of the configuration value is pushed on the stack.
 */
void
ocConfig(const Node * node)
{
	VALUE *	v1;
	VALUE *	v2;
	int	type;
	VALUE	tempValue;

	v1 = CalcValue(node->kids[0]);
	v2 = 0;

	if (node->count > 1)
		v2 = CalcValue(node->kids[1]);

	if (v1->v_type != V_STR)
	{
		math_error("Non-string for config");
		/*NOTREACHED*/
	}

	/*
	 * Get the configuration type from the string.
	 */
	type = configtype(v1->v_str->s_str);

	if (type < 0)
	{
		math_error("Unknown config name \"%s\"", v1->v_str->s_str);
		/*NOTREACHED*/
	}

	/*
	 * Get the old config value.
	 */
	config_value(conf, type, &tempValue);

	/*
	 * Set the new config value and remove its argument.
	 */
	if (node->count > 1)
	{
		setconfig(type, v2);

		PopValue();
	}

	/*
	 * Remove the first argument and push the old config value.
	 */
	PopValue();
	PushValue(&tempValue);
}
