/*
 * Copyright (C) 2009  David I. Bell
 *
 * Calc is open software; you can redistribute it and/or modify it under
 * the terms of the version 2.1 of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Calc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General
 * Public License for more details.
 *
 * A copy of version 2.1 of the GNU Lesser General Public License is
 * distributed with calc under the filename COPYING-LGPL.  You should have
 * received a copy with calc; if not, write to Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Node dumping module.
 *
 * This contains routines to describe the values of nodes for debugging.
 * This is basically for terminal nodes, since other nodes are usually
 * adequately described by their kids.
 *
 * Note: The string returned by one of these dump routines is only valid
 * until the next such call since it points into a static buffer.
 */
#include <stdio.h>

#include "calc.h"
#include "node.h"
#include "nodeinfo.h"
#include "userfunc.h"
#include "builtinfunc.h"
#include "method.h"
#include "symbol.h"


/*
 * Local storage for returned strings.
 */
static	char	dumpBuffer[256];

/*
 * Local routines.
 */
static const char * GetOwnerNodeName(const Node * node);


/*
 * Dump a function entry.
 */
const char *
odFunctionEntry(const Node * node)
{
	const char *	name;

	name = (const char *) node->data;

	if (name == 0)
		return "Null function name";

	sprintf(dumpBuffer, "\"%.100s\"", name);

	return dumpBuffer;
}


/*
 * Dump a label node.
 */
const char *
odLabel(const Node * node)
{
	if (node->data == 0)
		return "Null label name";

	sprintf(dumpBuffer, "%.100s", (const char *) node->data);

	return dumpBuffer;
}


/*
 * Dump a GOTO statement.
 */
const char *
odGotoStatement(const Node * node)
{
	if (node->data == 0)
		return "Null label name";

	sprintf(dumpBuffer, "%.100s (pop %ld)", (const char *) node->data,
		node->id);

	return dumpBuffer;
}


/*
 * Dump a BREAK statement.
 * This points to the node which it breaks.
 */
const char *
odBreakStatement(const Node * node)
{
	const Node *	ownerNode;
	const char *	ownerText;

	if (node->data == 0)
		return "Missing owner node";

	ownerNode = (const Node *) node->data;
	ownerText = GetOwnerNodeName(ownerNode);

	sprintf(dumpBuffer, "Owner %s, stack adjust %ld", ownerText, node->id);

	return dumpBuffer;
}


/*
 * Dump a CONTINUE statement.
 * This points to the node which it continues.
 */
const char *
odContinueStatement(const Node * node)
{
	const Node *	ownerNode;
	const char *	ownerText;

	if (node->data == 0)
		return "Missing owner node";

	ownerNode = (const Node *) node->data;
	ownerText = GetOwnerNodeName(ownerNode);

	sprintf(dumpBuffer, "Owner %s, stack adjust %ld", ownerText, node->id);

	return dumpBuffer;
}


/*
 * Dump a TRY statement.
 * This contains the id of a local variable to hold an error value
 * or -1 if there is no such variable.
 */
const char *
odTryStatement(const Node * node)
{
	if (node->id < 0)
		return "No catch variable";

	sprintf(dumpBuffer, "Id %ld", node->id);

	return dumpBuffer;
}


/*
 * Dump the QUIT statement.
 * This contains the id of a string to be printed or -1.
 */
const char *
odQuitStatement(const Node * node)
{
	const STRING *	string;

	if (node->id < 0)
		return "No message";

	string = findstring(node->id);

	if (string == 0)
		return "Null string";

	sprintf(dumpBuffer, "\"%.100s\"", string->s_str);

	return dumpBuffer;
}


/*
 * Dump the ASSERT statement.
 * This contains the id of a string to be printed or -1.
 */
const char *
odAssertStatement(const Node * node)
{
	const STRING *	string;

	if (node->id < 0)
		return "No message";

	string = findstring(node->id);

	if (string == 0)
		return "Null string";

	sprintf(dumpBuffer, "\"%.100s\"", string->s_str);

	return dumpBuffer;
}


/*
 * Dump a local variable name.
 */
const char *
odLocalAddr(const Node * node)
{
	sprintf(dumpBuffer, "Id %ld", node->id);

	return dumpBuffer;
}


/*
 * Dump a global variable name.
 */
const char *
odGlobalAddr(const Node * node)
{
	const GLOBAL *	global;

	global = (const GLOBAL *) node->data;

	if (global == 0)
		return "Null variable";

	sprintf(dumpBuffer, "%.100s", global->g_name);

	return dumpBuffer;
}


/*
 * Dump a parameter variable name.
 */
const char *
odParamAddr(const Node * node)
{
	sprintf(dumpBuffer, "Id %ld", node->id);

	return dumpBuffer;
}


/*
 * Dump an element address node.
 */
const char *
odElementAddr(const Node * node)
{
	sprintf(dumpBuffer, "%ld", node->id);

	return dumpBuffer;
}


/*
 * Dump an index address node.
 */
const char *
odIndexAddr(const Node * node)
{
	return ((node->id != 0) ? "write" : "read");
}


/*
 * Dump a literal integer value.
 */
const char *
odIntLiteral(const Node * node)
{
	sprintf(dumpBuffer, "%ld", node->id);

	return dumpBuffer;
}


/*
 * Dump a history value.
 */
const char *
odHistoryValue(const Node * node)
{
	sprintf(dumpBuffer, "$%ld", node->id);

	return dumpBuffer;
}


/*
 * Dump an imaginary constant.
 */
const char *
odImaginary(const Node * node)
{
	char *		cp;
	NUMBER *	q;

	q = constvalue(node->id);

	if (q == NULL)
	{
		sprintf(dumpBuffer, "Bad id %ld", node->id);

		return dumpBuffer;
	}

	math_divertio();
	fitprint(q, 32);
	cp = math_getdivertedio();

	sprintf(dumpBuffer, "%.100s i", cp);

	free(cp);

	return dumpBuffer;
}


/*
 * Dump an object creation.
 */
const char *
odObjectCreate(const Node * node)
{
	sprintf(dumpBuffer, "Id %ld", node->id);

	return dumpBuffer;
}


/*
 * Dump a string literal.
 */
const char *
odString(const Node * node)
{
	const STRING *	string;

	string = findstring(node->id);

	if (string == 0)
		return "Null string";

	sprintf(dumpBuffer, "\"%.100s\"", string->s_str);

	return dumpBuffer;
}


/*
 * Dump a number constant.
 */
const char *
odNumber(const Node * node)
{
	char *		cp;
	NUMBER *	q;

	q = constvalue(node->id);

	if (q == NULL)
	{
		sprintf(dumpBuffer, "Bad id %ld", node->id);

		return dumpBuffer;
	}

	math_divertio();
	fitprint(q, 32);
	cp = math_getdivertedio();

	strcpy(dumpBuffer, cp);

	free(cp);

	return dumpBuffer;
}


/*
 * Dump a method function call.
 */
const char *
odMethodCall(const Node * node)
{
	const char *	name;

	name = MethodName(node->id);

	if (*name)
		return name;
	
	sprintf(dumpBuffer, "Bad id %ld", node->id);

	return dumpBuffer;
}


/*
 * Dump a user-define function call.
 */
const char *
odUserCall(const Node * node)
{
	const UserFunc *	func;

	/*
	 * Find the function to be called.
	 */
	func = FindUserFunction(node->id);

	if (func)
		return func->name;

	sprintf(dumpBuffer, "Bad id %ld", node->id);

	return dumpBuffer;
}


/*
 * Dump a built-in function call.
 */
const char *
odBuiltinCall(const Node * node)
{
	const char *	name;

	name = BuiltinName(node->id);

	if (*name)
		return name;

	sprintf(dumpBuffer, "Bad id %ld", node->id);

	return dumpBuffer;
}


/*
 * Return a short owner opcode name for a break or continue node.
 */
static const char *
GetOwnerNodeName(const Node * node)
{
	if (node == 0)
		return "No owner node";

	if (node->info == 0)
		return "No owner node info";

	switch (node->info->opcode)
	{
		case OP_SWITCH_STATEMENT:
			return "switch";

		case OP_FOR_STATEMENT:
			return "for";

		case OP_WHILE_STATEMENT:
			return "while";

		case OP_DO_STATEMENT:
			return "do";

		default:
			return "Non-owner opcode";
	}
}
